import typer
import csv 
import time
from rich import print
from rich.console import Console
from rich.columns import Columns
from rich.table import Table
from rich.panel import Panel

from datetime import datetime

def abort(msg):
    print("[bold red]Error:[/bold red] "+msg)
    raise typer.Abort()

def now():
    return int(time.time())

def join(answers):

    filtered = []

    for x in answers:
        if answers[x] != None:
            to_append = answers[x]
            if type(answers[x]) is list:
                to_append = ",".join(answers[x])
                
            filtered.append(to_append)

    return "::".join(filtered)

def get_raw(source):    
    if source.type == 'file':        
        with open(source.path, newline='') as f:
            reader = csv.DictReader(f, skipinitialspace=True)
            headers = reader.fieldnames
            rows = []
            for row in reader:
                rows.append(row)
    else: 
        raise Exception("Source type 'api' is not yet implemented.")
    
    return { "headers": headers, "rows": rows}

def ts_to_str(ts):
    dt = datetime.fromtimestamp(ts)
    return dt.strftime("%c")

def render_state(state):
    match state:
        case 0:
            return "pending"
        case 1: 
            return "success"
        case _:
            return "failure"

def print_jobs(jobs):
    console = Console()
    header = ["Job Id", "Scheduled", "Created", "State"]
    table = Table(*header, expand=True, highlight=True, box=None, title_justify="left", show_lines="True")
    console.print()
    for job in jobs:
        row = []
        for key,val in job.items():
            if key == 'hash':
                row.append(val[:10])
            if key in ['created','scheduled']:
                row.append((ts_to_str(val)))
            if key == 'state':
                row.append(render_state(val))
        table.add_row(*row)
    console.print(table)
    console.print()


# def render_table(header, body):
#     console = Console()

#     if header:
#         table = Table(*header)

#     for row in body:
#         table.add_row(*row)
    
#     console.print(table)


from odk_mailer.lib import globals
from datetime import datetime
import time
import json
import hashlib
from enum import Enum
import os

class JobType(Enum):
    FILE = 'file'
    API = 'api'

class Source:
    type: str
    path: str
    project: int
    hostname: str
    username: str
    password: str

    def __init__(self, lst: []):
        if not len(lst) in [2,6]: 
            raise Exception("Source input must have length 2 or 6.")
        
        self.type = JobType(lst[0]).value
        self.path = lst[1]
        if len(lst) == 6:
            self.project = int(lst[2])
            self.hostname = lst[3]
            self.username = lst[4]
            self.password = lst[5]

class Fields:
    email: str
    data: []

    def __init__(self, lst: []):
        if not len(lst) in [1,2]:
            raise Exception("Fields input must have length 1 or 2.")
        
        self.email = lst[0]
        # data fields are optional
        if len(lst) == 2:
            self.data = lst[1].split(",")

class Message:
    sender: str
    format: str
    source: str
    content: str

    def __init__(self, lst: []):
        if not len(lst) == 4:
            raise Exception("Message input must have length 4.")
                
        self.sender = lst[0]
        self.format = lst[1]
        self.source = lst[2]
        self.content = lst[3]        

class Schedule:
    timestamp: int

    def __init__(self,int:int):
        self.timestamp = int

class Job:
    created: int
    source: Source
    fields: Fields
    message: Message
    schedule: Schedule
    recipients: []
    json: str
    hash: str

    def __init__(self, source:Source, fields:Fields, message:Message, schedule:Schedule, raw:{}):

        self.created = int(time.time())
        self.source = vars(source)
        self.fields = vars(fields)
        self.message = vars(message)
        self.schedule = schedule.timestamp

        self.setRecipients(raw)
        self.setJSON()
        self.setHash()

    def setRecipients(self, data):
        recipients = []
        for row in data["rows"]:
            f_row = {k: row[k] for k in self.fields["data"] + [self.fields["email"]]}
            recipients.append(f_row)
        self.recipients = recipients

    def setJSON(self):
        self.json = json.dumps(vars(self), ensure_ascii=True, indent=4)

    def setHash(self):
        self.hash = hashlib.sha256(self.json.encode()).hexdigest()

    # does two things: saves mailjob as <hash>.json 
    # and adds it as an entry to jobs.json with hash as id
    def save(self):

        # save files
        path_to_job = os.path.join(globals.odk_mailer_job, self.hash+'.json')
        with open(path_to_job, 'w', encoding='utf-8') as f:
            f.write(self.json)

        with open(globals.odk_mailer_jobs, "r+") as f:
            jobs = json.load(f)
            save_job = {
                "hash": self.hash,
                "scheduled": self.schedule, 
                "created": self.created,
                "state":0
            }
            jobs.append(save_job)
            f.seek(0)
            f.truncate()
            f.write(json.dumps(jobs))

        return save_job



import sys
import os
import json
from odk_mailer.lib import prompts, validators, utils, globals, smtp
from odk_mailer.classes.job import Job
from odk_mailer.classes.mailer import Mailer
from odk_mailer.classes.config import Config

def run(hash_or_id, dry=False, verbose=False):

    odk_mailer_config = Config()

    if not odk_mailer_config:
        utils.abort("Cannot send emails without Config")

    if not hash_or_id:
        utils.abort("ID/Hash is required")

    with open(globals.odk_mailer_jobs, "r") as f:
        jobs = json.load(f)

    found = next((obj for obj in jobs if obj["hash"].startswith(hash_or_id)), None)
    if not found:
        utils.abort("Job not found.")

    hash = found['hash']
     
    # check if ready to be sent
    # simple check: scheduled <= now
    # advan. check: hasReminders AND UnsentReminderTime <= now
    
    # if found["scheduled"] > utils.now():
    #     utils.abort("Schedule is in future")

    mailer = Mailer(hash, dry, verbose, odk_mailer_config)
    mailer.send()
    # in case we have a reminder case, generate reminder contents from reminders/hash_reminderId.json

    # update job state: pending, success, errors

def delete(hash):
    if not hash:
            utils.abort("ID is required")

    with open(globals.odk_mailer_jobs, "r") as f:
        jobs = json.load(f)

    found = next((obj for obj in jobs if obj["hash"].startswith(hash)), None)

    if not found:
        utils.abort("Job not found.")
   
    # tbd: confirm
    
    # deletion from /job/<hash>.json
    path_job = os.path.join(globals.odk_mailer_job, found['hash']+'.json')
    if os.path.exists(path_job):
        os.remove(path_job)

    # deletion from /jobs.json
    jobs_updated = list(filter(lambda x: x['hash']!=found['hash'], jobs))   
    
    with open(globals.odk_mailer_jobs, "w") as f:
        f.write(json.dumps(jobs_updated))

    print("Deleted " + found['hash'])


def create(source, fields, message, schedule):

    if not source:
        p_source = prompts.source()
        source  = utils.join(p_source)  # stringify answers
    
    v_source = validators.source(source)

    raw = utils.get_raw(v_source)

    if not fields:
        p_fields = prompts.fields(raw["headers"])
        fields = utils.join(p_fields)

    v_fields = validators.fields(fields, raw["headers"])

    if not message:
        p_message = prompts.message()
        message = utils.join(p_message)

    v_message = validators.message(message)

    if not schedule:
        p_schedule = prompts.schedule()
        if p_schedule["now"]:
            schedule = "now"
        else: schedule=p_schedule["future"]

    v_schedule = validators.schedule(schedule)

    # tbd: Reminders
    # reminders have two attributes:
    # total_amount, frequency, e.g. 3 times in total, every hour|day|week|custom frequency
    # after first scheduled send

    # reminders will be stored inside .odk-mailer/reminder/hash_instance.json
    # having updated recipients from non-respondents (calculated from base recipients and respondents) 
    # on a per reminder case

    # https://github.com/tertek/zapier-odk-scripts/blob/main/get_non_responding.py
    # reminders require api connection and following inputs
    # form_register = "test_form_register" #name of form that is used for registration, given as api or csv
    # form_follow = "test_form_follow"  # name of form that is used for follow up
    # # if use_form_attachment
    # form_follow_attached = "test_form_follow_attached" # name of form that is used for follow up; 
    # form_attachment = "follow.csv" # name of form attachment attached to form_follow; 
    # # field config
    # field_email_register = "email_register" # name of email field for registration form, given
    # field_email_follow = "email_follow"
    # field_email_follow_attach = "email"

    job = Job(v_source, v_fields, v_message, v_schedule, raw)
    
    if "pytest" in sys.modules:
        # testing
        print(vars(job))
        sys.exit()
        
    saved = job.save()

    print()
    print("Created " + saved["hash"])
    print()

    # validate recipients and process invalid emails in case
    # if not recipients.validate(email_field):
    #     utils.render_table(["id", "email_field", "error"], recipients.invalidEmails)
    #     ignore_invalid_emails = typer.confirm("Invalid emails found. Would you like to continue although you have invalid emails?")
    
    #     if not ignore_invalid_emails:
    #         raise typer.Exit("\nAborted.")
    # store mail-tasks in a text file or JSON https://www.w3schools.com/python/python_json.asp
    # https://stackoverflow.com/a/24608746/3127170
    # the task will be stored with final data

    # task = {
    #   csv_file: path/to/file.csv
    #   email_field: email
    #   message: <msg>
    #   sender: <sndr>
    #   reminders: ...
    # }
        
    # add cron-job via https://pypi.org/project/python-crontab/ and https://stackabuse.com/scheduling-jobs-with-python-crontab/
    # We will need a single job, that checks every hour if we have open jobs (with reminder tasks.)
    # Process reminder tasks as follows: 
    # 1. Perform API request to calculate reminder recipients
    # 2. Send Emails based on calculated recipients
    # 3. Summarize progress


    #
    # odk-mailer run command <hash>
    #

    # 1. check if job exists under <hash>
    # 2. get job details
    # 3. replace placeholder in message with data (python template engine)
    # 4. 

def list_jobs():

    with open(globals.odk_mailer_jobs, "r+") as f:
        jobs = json.load(f)

    utils.print_jobs(jobs)


def evaluate(dry=False):

    with open(globals.odk_mailer_jobs, "r+") as f:
        jobs = json.load(f)

    evals = []

    for job in jobs:
        if job["scheduled"] <= utils.now():
            # simple evaluation: check if scheduled time is smaller/equal to now
            # if true, add to selected list
            evals.append(job["hash"])

        ###
        # untested code, since reminders are not yet implemented in create command
        elif "reminders" in job and len(job["reminders"]) > 0:
            # advanced evaluation: addtionally check if job has reminder times that are smaller/equal to now
            # if true, add to selected list
            print("Addiitonally checking if we have any valid reminders")
            for reminder in job["reminders"]:
                if reminder["timestamp"] <= utils.now():
                    evals.append(job["hash"])
                    break
        ###
                
        else:
            # skipping this job since not qualified to be run
            pass

    if dry:
        print(evals)
        print(len(evals))

    else:
        for eval in evals:
            run(eval)

def test(sender, recipient, host, port):
    print()
    print(f"Sending test mail from {sender} to:  {recipient} via: {host}:{port}")
    print()
    smtp.send_mail(sender, recipient, host, port)


import typer
import csv
from email_validator import validate_email, EmailNotValidError

# return unformatted string instead of raising error
# when key is missing within dictionary
# https://stackoverflow.com/a/17215533/3127170
class SafeDict(dict):
    def __missing__(self, key):
        return '{' + key + '}'

class Recipients:
    def __init__(self, path):
        self.path  = path
        self.fieldnames = []
        self.data = []

        self.isValid = False
        self.invalidEmails = []

        # Initialize 
        self.init()

    def init(self):

        # # invalid file extension
        # ext = os.path.splitext(self.path)[-1].lower()
        # if not ext == ".csv":
        #     raise typer.Exit("Invalid file extension.")

        # # invalid file path
        # if not os.path.isfile(self.path):
        #     raise typer.Exit("Invalid file path.")

        # read data
        with open(self.path, newline='') as f:
            reader = csv.DictReader(f, skipinitialspace=True)
            self.fieldnames = reader.fieldnames
            for row in reader:
                self.data.append(row)

    def validate(self, email_field):
        invalid_emails = []
        total = 0

        print("\nValidating recipients...")
        with typer.progressbar(self.data) as progress:
            for row in progress:
                total += 1
                email = row[email_field]
                try:
                    if not email:
                        raise EmailNotValidError("Email address missing. Check for missing delimiters (',') in your CSV file.")
                    # Disable DNS checks since this can be blocked for unknown reasons within network.
                    validate_email(email, check_deliverability=True)
                except EmailNotValidError as e:
                    invalid = [str(total) ,email, str(e)]
                    invalid_emails.append(invalid)
                
        print(f"Validated {total} entries.\n")

        self.invalidEmails = invalid_emails
        self.numEmails = len(self.data)
        
        if len(invalid_emails) == 0:
            self.isValid = True

        return self.isValid

    # def test_filling(self):
    #     print(self.fieldnames)
    #     print(f'file path is {self.path}')
        
    #     txt = "The firstname is {firstname}\n lastname is {lastname} \n email is {email} \n not filled {not_filled}"
    #     for row in self.data:
    #         print(txt.format_map(SafeDict(row)))


import inquirer
from odk_mailer.lib import validators

def source():
    questions = [
        inquirer.List('type', 
                    message="Select source type:",
                    choices= [
                        ("CSV file.", "file"),
                        ("ODK API", "api")
                    ],                      
                    carousel=True
        ),
        inquirer.Path('file_path',
                    message="Input local path to CSV file",
                    path_type=inquirer.Path.FILE,
                    exists=True,
                    normalize_to_absolute_path=True,
                    ignore=lambda x: x["type"] == "api"
        ),
        inquirer.Text('api_form',
                    message="Input form/attachment name",                      
                    ignore=lambda x: x["type"] == "file"
                    #tbd: add regex valdation
        ),
        inquirer.Text('api_proj',
                    message="Input project id",
                    validate= validators.int_only,
                    ignore=lambda x: x["type"] == "file"
        ),
        inquirer.Text('api_host',
                    message="Input ODK host url",
                    ignore=lambda x: x["type"] == "file",
                    # tbd validate url, https
        ),
        inquirer.Text('api_user',
                    message="Input ODK username",
                    ignore=lambda x: x["type"] == "file"
        ),
        inquirer.Password('api_pass',
                    message="Input ODK password",
                    ignore=lambda x: x["type"] == "file"
        )
    ]
    return inquirer.prompt(questions, raise_keyboard_interrupt=True)

def fields(headers=[]):
    questions = [
        inquirer.List('email',
                    message="Select email field",
                    choices=headers,
                    default='email' if 'email' in headers else "",
                    carousel=True
        ),
        inquirer.Checkbox("data", 
                    message="Select data field(s)",
                    choices= lambda x: filter(lambda y: y != x["email"], headers)
        )        
    ]
    return inquirer.prompt(questions, raise_keyboard_interrupt=True)

def message():
    questions = [
        inquirer.Text('sender',
                    message="Input message sender",
                    default="odk@swisstph.ch",  # remove
                    validate=validators.email_address                      
        ),
        inquirer.List('format', 
                    message="Select message format:",
                    choices= [
                        ("Text", "txt"),
                         ("HTML", "html")
                    ],                      
                    carousel=True
        ),
        inquirer.List('source', 
                    message="Select message source:",
                    choices= [
                        ("Input as string", "stdin"),
                        ("File from path", "path")
                    ],                      
                    carousel=True
        ),        
        inquirer.Text('content_stdin',
                    message="Input message string",
                    ignore=lambda x: x["source"] != "stdin",
                    validate= validators.not_empty
        ),
        inquirer.Path('content_path',
                    message="Message file path",
                    path_type=inquirer.Path.FILE,
                    exists=True,
                    normalize_to_absolute_path=True,
                    ignore=lambda x: x["source"] != "path"
                    
        )
    ]
    return inquirer.prompt(questions, raise_keyboard_interrupt=True)

def schedule():
    questions = [
        inquirer.List('now',
                    message="Send immediately now or schedule for later?",
                    choices=[
                        ("Send now",True),
                        ("Schedule for later", False)
                    ]
        ),
        inquirer.Text('future',
                    message="Input schedule time and date",
                    ignore=lambda x: x["now"] == True,
                    validate= validators.date_format
        )
    ]

    return inquirer.prompt(questions, raise_keyboard_interrupt=True)

def confirm():
    return inquirer.prompt(
        [
            inquirer.Confirm("confirm", message="Create MailJob?")
        ], 
        raise_keyboard_interrupt=True)


from inquirer import errors
from odk_mailer.lib import utils
from odk_mailer.classes.job import Source, Fields, Message, Schedule
import re
import os
from datetime import datetime
import time

def source(s_str: str):

    s_lst = s_str.split("::", 6)
    s_len = len(s_lst)

    if not s_len in [2,6]:        
        utils.abort("Invalid file/api source string")

    if s_len == 2:
        s_type = s_lst[0]
        s_path = s_lst[1]

        if s_type != "file":
            utils.abort("Invalid source type, expected 'file'.")
        ext = os.path.splitext(s_path)[-1].lower()
        # invalid file extension
        if not ext == ".csv":
            utils.abort("Invalid file extension, expected '.csv'")
        # invalid file path
        if not os.path.isfile(s_path):
            utils.abort("Invalid file path, expected existing file")
        
        source = Source(s_lst)

    #tbd        
    elif s_len == 6:
        s_type = s_lst[0]
        s_path = s_lst[1] # odk form/attachment id
        s_proj = s_lst[2] # odk project id
        s_host = s_lst[3] # odk host url
        s_user = s_lst[4] # odk username
        s_pass = s_lst[5] # odk password

        if s_type != "api":
            utils.abort("Invalid source type, expected 'api'.")
        # check API connection & credentials
        # promptAPIcredentials() // host, user, pass, project
        # checkAPIconnection()  // auth endpoint

        source = Source(s_lst)

    return source

def fields(f_str:str, headers=[]):

    f_lst = f_str.split("::", 2)
    f_len = len(f_lst)

    if not f_len in [1,2]:        
        utils.abort("Invalid email/data fields string")
    
    f_email = f_lst[0]

    if f_email not in headers:
        utils.abort(f"Invalid email field '{f_email}'.")

    if f_len == 2:
        f_data = f_lst[1].split(",") 
        for field in f_data:
            if not field in filter(lambda x: x != f_email, headers):
                utils.abort(f"Invalid data field '{field}'")
    
    return Fields(f_lst)

def message(m_str: str):

    m_lst = m_str.split("::", 4)
    m_len = len(m_lst)

    if not m_len == 4:
        utils.abort("Invalid message string, expected [sender]::[format]::[source]::[content]")

    m_sender = m_lst[0]
    m_format = m_lst[1]
    m_source = m_lst[2]
    m_content = m_lst[3]

    # validate message sender
    if not re.match(r"^\S+@\S+\.\S+$", m_sender) :
        raise utils.abort("Invalid message sender. Use valid email address.")
    
    if m_format not in ["txt", "html"]:
        raise utils.abort("Invalid message format. Use either 'txt' or 'html.")
    
    if m_source not in ["stdin", "path"]:
        raise utils.abort("Invalid message source. Use either 'stdin' or 'path.")
    
    # tbd: validate m_content (check if file exists and/or content is valid h)
    
    return Message(m_lst)

def schedule(str: str):
    if str == "now":
        schedule = int(time.time())
    else: 
        # tbd: date format validation
        _datetime = datetime.fromisoformat(str)
        schedule = int(datetime.timestamp(_datetime))

    return Schedule(schedule)

def int_only(_, current):
    re_integer = "^[0-9]*$"
    if not re.search(re_integer, current):
        raise errors.ValidationError('', 'Invalid project id format. Has to be a number')
    return True

def date_format(_, current):
    # tbd: must be future
    re_YYYY_MM_DD_hh_mm = "[0-9]{4}-(0[1-9]|1[0-2])-(0[1-9]|[1-2][0-9]|3[0-1]) (2[0-3]|[01][0-9]):[0-5][0-9]"
    if not re.search(re_YYYY_MM_DD_hh_mm, current):
        raise errors.ValidationError('', 'Invalid date format. Has to be YYYY-MM-DD hh:mm')       
    return True

def email_address(_, current):
    re_email_address  = "^\S+@\S+\.\S+$"
    if not re.search(re_email_address, current):
        raise errors.ValidationError('', 'Invalid email address format.')
    return True

def not_empty(_, current):
    if current == "":
         raise errors.ValidationError('', 'Content cannot be empty.')
    return True